/*
** Module   : hql.g
** Abstract : ANTLR grammar for a lexer and parser which understands HQL
**            where clause syntax, including a P2J-specific extension thereof
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060314   @25235 Created initial version. An ANTLR grammar for
**                           an HQL where clause lexer and parser.
** 002 ECF 20060808   @28574 Fixed defect in function parameter parsing.
**                           Drop COMMA from AST.
** 003 ECF 20060815   @28690 Added support for ternary expressions. These
**                           occur as CASE WHEN ... THEN ... ELSE ... END
**                           statements.
** 004 ECF 20060920   @29794 Added support for string concatentation. Uses
**                           the binary string concatenation operator
**                           ('||').
** 005 ECF 20070327   @32616 Fixed STRING token rule. Must allow embedded,
**                           escaped, single quote (''). Previously, we
**                           were removing the leading single quote in
**                           such a pair, which improperly removed the
**                           escape character HQL/SQL needs to parse this
**                           correctly.
** 006 ECF 20070502   @33372 Added limited support for subselect phrases.
**                           Only those features required to support the
**                           server-side implementation of converted,
**                           nested CAN-FIND statements are implemented.
** 007 ECF 20071216   @36493 Permit 'as' subclause of a 'from' clause to
**                           be optional. These are not necessary in some
**                           subselect phrases.
** 008 ECF 20080311   @37488 Fixed defect within alias rule. Unqualified
**                           property references were being saved as ALIAS
**                           nodes.
** 009 SVL 20090810   @43576 Added CAST token.
** 010 SVL 20110622          Maintain newlines into string literals.
** 011 OM  20130501          Added LONG_LITERAL token.
*/

header
{
/*
** Module   : HQLParser.java
**            HQLParserTokenTypes.java
**            HQLLexer.java
** Abstract : 
**
** Copyright (c) 2006, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20060314 ADD   @25235 WARNING, THIS IS A GENERATED FILE!!!
**                               DO NOT EDIT THIS FILE. The original source
**                               file is hql.g!
*/

package com.goldencode.p2j.persist.hql;

import java.util.*;
import java.util.List;
import java.text.*;
import java.io.*;
import antlr.*;
import com.goldencode.p2j.uast.*;
}

/**
 * Creates an Abstract Syntax Tree (AST) representation of an HQL expression
 * in infix notation from an input stream of tokens (provided by the
 * {@link HQLLexer}).  More specifically, the parser does the following:
 * <ul>
 *    <li> Drives the lexer by calling its' <code>nextToken</code> method.
 *         This method tokenizes the lexer's input stream of characters and
 *         returns the next found <code>Token</code> object.
 *    <li> Provides the {@link #expression} method as an entry point.
 *    <li> Builds an AST (&quot;intermediate form representation&quot;) of
 *         all the recognized tokens. This AST is designed for subsequent
 *         processing by the {@link com.goldencode.p2j.persist.HQLPreprocessor
 *         HQLPreprocessor}.
 *    <li> Handles the context sensitive aspects of symbol resolution.
 *    <li> To the extent that is possible while still generating a well
 *         structured tree, the parser has been structured to enforce
 *         syntactic correctness.
 * </ul>
 * <p>
 * This grammar is highly specific to ANTLR 2.7.4 and it may not be valid
 * in future versions of ANTLR.  This is due to the fact that the generated
 * code and the grammar syntax of ANTLR can change between versions.  These
 * factors are the most important constraints defining the structure of the
 * resulting grammar.  As a result, it is possible that the grammar will
 * require changes to run in future versions of ANTLR.
 * <p> 
 * <b>Please note that this is a generated file using ANTLR 2.7.4 and a
 * grammar specified in <code>hql.g</code>.</b>
 *
 * @author ECF
 * @author GES
 */
class HQLParser extends Parser;

options
{   
   k = 3;
   buildAST = true;
   ASTLabelType = "HQLAst";
}

tokens
{
   BOOL_FALSE;
   BOOL_TRUE;
   NULL;
   UN_MINUS;
   IS_NULL;
   NOT_NULL;
   DEC_LITERAL;
   OR;
   AND;
   NOT;
   DOT;
   IS;
   ALIAS;
   FUNCTION;
   PROPERTY;
   SUBSCRIPT;
   INDEX;
   CASE;
   WHEN;
   THEN;
   ELSE;
   END;
   TERNARY;
   SUBSELECT;
   SELECT;
   FROM;
   WHERE;
   AS;
   DMO;
   CAST;
}
   
// Custom instance variables, constructors, methods...
{
   /** Name of the specialized AST class to use when parsing. */
   private static final String AST_CLASS = HQLAst.class.getName();
   
   /** Provides an efficient reverse lookup of names to token types. */
   private static Map tokenLookup = null;
   
   /** Next index value to be assigned to a query substitution node */
   private int nextSubstitutionIndex = 0;
   
   /**
    * Simple walker to print out the contents of an AST for debugging.
    *
    * @param    ast
    *           The current node to print.
    * @param    level
    *           The number of levels down from the root (0 == root node),
    *           which determines the indention level to print.
    */
   public static void visit(CommonAST ast, int level)
   {
      int          spaces = level * 3;
      StringBuffer sb     = new StringBuffer();
      
      // generate the indent based on the level from the root
      for (int i = 0; i < spaces; i++)
      {
         sb.append(' ');
      }
      
      // add our current node's text
      sb.append("[ ");
      sb.append(ast.getText());
      sb.append(" ] <");
      sb.append(_tokenNames[ast.getType()]);
      sb.append(">");
      
      // print out the current node
      System.out.println(sb.toString());
   
      CommonAST child = (CommonAST) ast.getFirstChild();
      level++;
      
      // now visit all children
      while (child != null)
      {
         visit(child, level);
         child = (CommonAST) child.getNextSibling();
      }
   }
   
   /**
    * Translates a text representation of a parser token type into the
    * actual integer value.
    *
    * @param    tokenName
    *           The text representation of the parser token type.
    *
    * @return   A valid parser-defined integer token type value or -1 if
    *           no match was found.
    */
   public static int lookupTokenType(String tokenName)
   {
      // initialize map if necessary
      if (tokenLookup == null)
      {
         int len = _tokenNames.length;
         
         tokenLookup = new HashMap();
         
         // load the map with our lowercased symbolic name as the key and
         // the wrappered int token type as the value
         for ( int i = 0; i < len; i++ )
         {
            tokenLookup.put(_tokenNames[i].toLowerCase(), new Integer(i));
         }
      }
      
      Integer result = (Integer) tokenLookup.get(tokenName.toLowerCase());
      
      return (result == null ? -1 : result.intValue());
   }
   
   /**
    * Translates an integer value of a parser token type into the associated 
    * human readable text representation.
    *
    * @param    type
    *           The integer token type.
    *
    * @return   A valid parser-defined symbolic name or <code>null</code> if
    *           no match was found.
    */
   public static String lookupTokenName(int type)
   {
      if (type < 0 || type >= _tokenNames.length)
      {
         return null;
      }
      
      return _tokenNames[type];
   }
   
   /**
    * Implements special processing for <code>NULL</code> equality and
    * inequality tests.  If a <code>NULL</code> token is detected on either
    * side of either of the operators, the result will be rooted at an
    * <code>IS_NULL</code> (equality) or <code>NOT_NULL</code> (inequality)
    * node.  The operand that is <code>NULL</code> will be dropped and the
    * other operand will be the only child.    
    * <p>
    * This is separated from the main processing and really is a kind of
    * post-processing of the entire tree.
    */
   public void convertNullTest(CommonAST ast)
   {
      int type = ast.getType();
      
      // process the current node, only equality and inequality are of
      // interest
      if (type == EQUALS || type == NOT_EQ)
      {
         CommonAST first  = (CommonAST) ast.getFirstChild();
         CommonAST second = (CommonAST) first.getNextSibling();
         
         int ftype = first.getType(); 
         int stype = second.getType();
      
         // detect comparisons with NULL and rewrite the sub-expression
         if (ftype == NULL || stype == NULL)
         {
            // special case that can be optimized easily
            if (ftype == NULL && stype == NULL)
            {
               ast.setType(type == EQUALS ? BOOL_TRUE : BOOL_FALSE);
               ast.removeChildren();
            }
            else
            {
               // normal case: rewrite the token as our artificial unary
               // operator
               ast.setType(type == EQUALS ? IS_NULL : NOT_NULL);
               
               // only leave the non-NULL child behind as the single operand
               if (first.getType() == NULL)
               {
                  ast.setFirstChild(second);
               }
               else
               {
                  first.setNextSibling(null);
               }
            }
         }
      }
      
      // either way, process all children using a recursive call 
      CommonAST child = (CommonAST) ast.getFirstChild();
      
      while (child != null)
      {
         convertNullTest(child);
         child = (CommonAST) child.getNextSibling();
      }
   }
   
   /**
    * Parser error-reporting function.
    *
    * TODO:  reimplement to use logging...
    */
   public void reportError(RecognitionException exc)
   {
      super.reportError(exc);
   }
   
   /**
    * Provides a command line interface for an end user to drive and/or test
    * the HQLParser class.
    * <p>
    * Syntax:
    * <pre>
    *    java HQLParser &lt;expression&gt;
    * </pre>
    *
    * @param   args
    *          List of command line arguments.
    */
   public static void main(String[] args)
   {
      if (args.length != 1)
      {
         System.err.println("Syntax: java HQLParser <expression>");
         System.exit(-1);
      }
         
      try
      {
         StringReader     expr       = new StringReader(args[0]);
         HQLLexer         lexer      = new HQLLexer(expr);
         HQLParser        parser     = new HQLParser(lexer);
         
         BaseAST.setVerboseStringConversion(true, _tokenNames);
         
         parser.expression();         
         
         HQLAst result = (HQLAst) parser.getAST();
         
         result.fixups(null, null);
         
         // print out report
         visit(result, 0);
      }
       
      catch(Exception excpt)
      {
         excpt.printStackTrace();
      }
   }
}

/**
 * Parses an HQL expression and creates an AST that is suitable for further
 * processing.  This is the main "entry point" for HQL expression parsing.
 * The AST node class is set here (to {@link HQLAst}).
 * <p>
 * The AST is structured with operator nodes as the root nodes and each
 * operator (unary or binary) has the matching number of child nodes (1 or
 * 2 respectively).  Each child node is an operand and the semantics of
 * left and right operands (for binary operators) is maintained.
 * <p>
 * Non-operator nodes in the tree can be literals or HQL functions.  The list
 * of literals can be seen in the {@link #literal} rule.
 * <p>
 * All of the following operators are supported and the precedence order
 * follows that of SQL, which we assume is the same precedence order used by
 * HQL (TODO:  confirm this with Hibernate sources).
 * <pre>
 * lowest  logical OR              :  or
 *    |    conditional AND         :  and
 *    |    equality                :  =
 *    |    relational              :  <, >, <=, >=
 *    |    additive                :  +, -
 *    |    multiplicative          :  *, /
 *    |    unary operators         :  -, not
 *    |    function call           :  function_name(...)
 * highest parentheses             :  (...)
 * </pre>
 * <p>
 * To handle the precedence issue, each precedence level (see the table
 * above) is handled in a separate rule.  Starting at in this top level
 * entry point, the lowest precedence rule is referenced. This reference is
 * made with an optional <code>OR</code> operator.  If this operator is
 * present, it is the root of the resulting tree.  This means it is the
 * last operator to be evaluated since its operands will be processed
 * first.  This is the proper definition of lowest precedence.  If this
 * operator is present, then a second operand is expected.
 * <p>
 * Each precedence level is similarly constructed, the left operand being
 * required, all operators being optional and each operator choice an equal 
 * alternative. If any of the operators exists, then it is the root of the
 * subtree and a right operand is expected.  This holds true for all rules
 * that process binary operators.  Please note that this left-mandatory
 * and optional right side (operator + right operand) is critical for
 * allowing an expression to be as simple as a single variable, function
 * call or literal. Alternatively, it can expand into a highly nested set of
 * recursively parsed expressions that all evaluate to a single scalar result.
 * <p>
 * At the bottom of the hierarchy of rules are unary operators, these
 * are implemented as an optional prefix operator and a mandatory right
 * operand.
 * <p>
 * The final level is the primary expression.  This is where the precedence
 * operator parenthesis is handled.  It also is where we allow a single
 * literal, alias, or function reference.
 * <p>
 * The {@link #expr} rule is called as the only rule reference.
 * <p>
 */
expression
   :
      {
         setASTNodeClass(AST_CLASS);
      }
      (
         e:expr
      )
   ;

/**
 * Implements the lowest (1st) precedence level of expressions:
 * <code>OR</code> (the logical OR operator).  Valid expressions can contain
 * sub-expressions and this entry point is properly recursive without causing
 * any ambiguity.
 * <p>
 * This rule is designed to allow recursion (the highest precedence levels
 * {@link #function} and {@link #primary_expr} reference this rule again,
 * allowing infinite nesting of sub-expressions, with proper parsing,
 * evaluation and tree building.
 * <p>
 * This method is called by the top level {@link #expression} rule.  This
 * method calls the {@link #log_and_expr} rule.
 * <p>
 * Implements special processing for <code>NULL</code> equality and inequality
 * tests.  If a <code>NULL</code> token is detected on either side of either
 * of the operators, the result will be rooted at an <code>IS_NULL</code>
 * (equality) or <code>NOT_NULL</code> (inequality) node.  The operand that
 * is <code>NULL</code> will be dropped and the other operand will be the
 * only child.  This must be implemented after the entire tree is created 
 * and the method {@link #convertNullTest} is used to implement this
 * processing.
 * (TODO:  is this necessary?)
 */
expr
   :
      (
         log_and_expr
         (
            options { generateAmbigWarnings = false; }
            :
            OR^ log_and_expr
         )*
      )
      {
         // convert all equality and inequality tests against the null
         // literal into a special operator
         convertNullTest((CommonAST) ##);
      }
   ;
   
/**
 * Implements the 2nd lowest precedence level of expressions: <code>AND</code>
 * (the logical AND operator).
 * <p>
 * This is only called by the {@link #expr} rule.  This method calls the
 * {@link #equality_expr} rule.
 */
log_and_expr
   :
      equality_expr
      (
         options { generateAmbigWarnings = false; }
         :
         AND^ equality_expr
      )*
   ;
   
/**
 * Implements the 3rd lowest precedence level of expressions - the equality
 * operators.
 * <p>
 * This is only called by the {@link #log_and_expr} rule.  This method
 * calls the {@link #compare_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    =        (equality)
 *    !=       (inequality)
 * </pre>
 */
equality_expr
   :
      compare_expr
      (
           options { generateAmbigWarnings = false; }
           :
           ((EQUALS^ | NOT_EQ^) compare_expr)
         | i:IS^   { #i.setType(IS_NULL); }
           (NOT!   { #i.setType(NOT_NULL); })?
           NULL!
      )*
   ;
   
/**
 * Implements the 4th lowest precedence level of expressions - the comparison
 * operators.
 * <p>
 * This is only called by the {@link #equality_expr} rule. This method calls
 * the {@link #concat_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    >    (greater than)
 *    <    (less than)
 *    >=   (greater than or equal to)
 *    <=   (less than or equal to)
 *    like
 * </pre>
 */
compare_expr
   :
      concat_expr
      (
         options { generateAmbigWarnings = false; }
         :
         (GT^ | LT^ | GTE^ | LTE^ | LIKE^) concat_expr
      )*
   ;
   
/**
 * Implements the 5th lowest precedence level of expressions - concatenation.
 * <p>
 * This is only called by the {@link #compare_expr} rule. This method calls
 * the {@link #sum_expr} rule.
 * <p>
 * Provides the following operator:
 * <pre>
 *    ||   (concatenation of two string operands)
 * </pre>
 */
concat_expr
   :
      sum_expr
      (
         options { generateAmbigWarnings = false; }
         :
         CONCAT^ sum_expr
      )*
   ;
   
/**
 * Implements the 5th lowest precedence level of expressions - the additive
 * operators.
 * <p>
 * This is only called by the {@link #compare_expr} rule. This method calls
 * the {@link #prod_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    + (binary plus)
 *    - (binary minus)
 * </pre>
 */
sum_expr
   :
      prod_expr
      (
         options { generateAmbigWarnings = false; }
         :
         (PLUS^ | MINUS^) prod_expr
      )*
   ;

/**
 * Implements the 6th lowest precedence level of expressions - the
 * multiplicative operators.
 * <p>
 * This is only called by the {@link #sum_expr} rule. This method calls
 * the {@link #un_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    * (multiply)
 *    / (divide)
 * </pre>
 */
prod_expr
   :
      un_expr
      (
         options { generateAmbigWarnings = false; }
         :
         (MULTIPLY^ | DIVIDE^) un_expr
      )* 
   ;
   
/**
 * Implements the 7th lowest precedence level of expressions - the unary
 * operators.
 * <p>
 * This is only called by the {@link #prod_expr} rule. This method calls
 * the {@link #primary_expr} rule.
 * <p>
 * Provides the following operators:
 * <pre>
 *    -                  unary minus
 *    not                logical NOT
 * </pre>
 */
un_expr
   :
      {
         String cls = null;
      }
      (
         (
              m:MINUS^ { #m.setType( UN_MINUS ); }
            | NOT^
         )?
         primary_expr
      )
   ;
   
/**
 * Implements the 8th and highest precendence level of expressions, the
 * parenthesis precedence operators, {@link #literal}, {@link #function},
 * {@link #ternary}, {@link #property}, {@link #subselect}, and {@link
 * #parentheses} rules.
 * <p>
 * This rule is different in structure from the previous levels, as the
 * there is just a set of alternatives and one of them allows direct recursion
 * back to the <code>expr</code> expression entry point.
 * <p>
 * This is only called by the {@link #un_expr} rule.
 * <p>
 * See the rule {@link #expr} for more details.
 */
primary_expr
   :
      (
           // literal
           literal
           
         | // HQL function
           function
           
         | // Ternary using the CASE WHEN... construct
           ternary
           
         | // DMO property, possibly qualified by an alias
           property
           
         | // subselect expression
           subselect
           
         | // parenthesized sub-expressions (the precedence operator)
           parentheses
      )
   ;
   
/**
 * Matches HQL function signatures:  a SYMBOL followed by a parameter list in
 * parentheses.  The list may contain zero or more parameters, separated by
 * commas.  This rule is called by the {@link #primary_expr} rule.
 */
function
   :
      (
         s:SYMBOL^ LPARENS! ( expr ( COMMA! expr )* )? RPARENS!
         { #s.setType(FUNCTION); }
      )
   ;
   
/**
 * Matches a ternary expression which uses the CASE WHEN construct in its
 * ternary form.
 */
ternary
   :
      (
         CASE! w:WHEN^ expr THEN! expr ELSE! expr END!
         { #w.setType(TERNARY); }
      )
   ;
   
/**
 * Matches an alias for a data model object, which is simply a SYMBOL,
 * optionally followed by a property.  This rule is called by the {@link
 * #primary_expr} rule, but only after the {@link #function} rule is invoked.
 * This order is necessary to avoid ambiguity with the latter rule, which
 * looks for a leading SYMBOL token.
 */
alias
   :
      (
         s:SYMBOL^         { #s.setType(ALIAS); }
         ( property )?
      )
   ;
   
/**
 * Matches a qualified or unqualified property name, with an optional {@link
 * #subscript}.  This rule is called by the {@link #primary_expr} rule, but
 * only after the {@link #function} rule is invoked.  This order is necessary
 * to avoid ambiguity with the latter rule, which looks for a leading SYMBOL
 * token.
 */
property
   :
      (
         (s:SYMBOL^ DOT!)?
         {
            if (#s != null)
            {
               #s.setType(ALIAS);
            }
         }
         unqprop
      )
   ;
   
/**
 * Matches an unqualified property name, with an optional {@link #subscript}.
 */
unqprop
   :
      (p:SYMBOL^ (subscript)?)   { #p.setType(PROPERTY); }
   ;
   
/**
 * Matches a {@link #property} subscript, for array-type properties.
 */
subscript
   :
      (
         LBRACKET^
         (
              n:NUM_LITERAL
            | substitution
         )
         RBRACKET!
      )
   ;
   
/**
 * Matches a subselect phrase embedded within a where clause.  Called by
 * {@link #primary_expr} rule.
 */
subselect
   :
      {
         astFactory.makeASTRoot(currentAST, #[SUBSELECT, "subselect"]);
      }
      ((select_expr)? from_expr (where_expr)?)
   ;
   
/**
 * Matches a select phrase embedded within a where clause as part of a
 * subselect statement.  Called by {@link #subselect} rule.
 */
select_expr
   :
      SELECT^ function
   ;
   
/**
 * Matches a from phrase embedded within a where clause as part of a
 * subselect statement.  Called by {@link #subselect} rule.
 */
from_expr
   :
      FROM^
      t:SYMBOL    { #t.setType(DMO); }
      (as_expr)?
   ;

/**
 * Matches an as phrase, which is an optional assignment of an alias to a DMO
 * at the end of a from phrase.  Called by {@link #from_expr} rule.  This
 * portion of the phrase is optional if no other part of the subselect phrase
 * references an alias for the DMO.
 */
as_expr
   :
      AS!
      a:SYMBOL    { #a.setType(ALIAS); }
   ;
   
/**
 * Matches a where phrase embedded within a where clause as part of a
 * subselect statement.  Called by {@link #subselect} rule.  Recursively
 * calls {@link #expr} rule.
 */
where_expr
   :
      WHERE^ expr
   ;
   
/**
 * Matches a subexpression enclosed in parenetheses.  Called by {@link
 * #primary_expr} rule;  recursively calls {@link #expr} rule.
 */
parentheses
   :
      ( LPARENS^ expr RPARENS! )
   ;
   
/**  
 * Matches all variations of expression literals (constants).  This includes:
 * <pre>
 *    NUM_LITERAL
 *    LONG_LITERAL
 *    DEC_LITERAL
 *    STRING
 *    BOOL_TRUE
 *    BOOL_FALSE
 *    NULL
 *    SUBST
 * </pre>
 */
literal
   :
      (
           NUM_LITERAL
         | LONG_LITERAL
         | DEC_LITERAL
         | STRING
         | BOOL_TRUE
         | BOOL_FALSE 
         | NULL
         | substitution
      )
   ;
   
/**
 * Matches a query substitution parameter placeholder.  The "index" annotation
 * is written to this node;  its value is the index of this placeholder among
 * all such placeholders encountered in a left to right traversal of the where
 * clause.
 */
substitution
   :
      (
         s:SUBST
         {
            // Assign an index to this substitution parameter.
            #s.putAnnotation("index", new Long(nextSubstitutionIndex++));
         }
      )
   ;
   
/**
 * Tokenizes an HQL query statement (input as a stream of characters) into a
 * stream of tokens suitable for the {@link HQLParser}.
 * <p>
 * Each token has an integer token type by which the parser references and
 * matches tokens.  This is done to make a high performance parser. To make
 * this work, the lexer must create a token object when it matches a top
 * level rule, this token object includes the token type and the actual text
 * found in the original character stream.  The lexer's main job is to
 * look ahead into the character stream, switch into the top level rule that
 * matches some set of characters and when this is complete, to create a
 * valid token object and return this to the parser.  The lexer thus
 * implements a token stream interface from the parser's perspective.
 * <p>
 * <b>Please note that this is a generated file using ANTLR 2.7.4 and a
 * grammar specified in <code>hql.g</code>.</b>
 * <p>
 * Symbol resolution is the process by which the proper token type is  
 * assigned to a token created from a match to the generic symbol rule.
 * In the lexer, the only symbol resolution that can be done is to match
 * with reserved keywords.  Reserved keywords are given preference over 
 * other symbol types in any case where there is a conflict.
 * <p>
 * Other key design points:
 * <ul>
 *    <li> This grammar has been designed to ignore case.  All symbol
 *         definitions use lowercase characters and ANTLR generates code to
 *         match both the uppercase and lowercase version of each character.
 *         This is enabled using the lexer options section.
 *    <li> The lexer is defined to create a single token of any string.  All
 *         logic for matching string literals is embedded in the lexer.
 *    <li> At the top level of the lexer, all tokens are one of the
 *         following types: 
 *       <ul>
 *          <li> string literals
 *          <li> whitespace (spaces, tabs, carriage returns, line feeds)
 *          <li> operators
 *          <li> integer and decimal literals
 *          <li> reserved keywords 
 *          <li> user-defined symbols
 *       </ul>
 *    <li> The lexer's top level entry point is <code>nextToken</code>. This
 *         method has a for-loop that uses up to 3 characters of lookahead
 *         to identify which of the top level token rules to call. To do
 *         this, each token rule's left-most match characters are "rolled up"
 *         and tested in this top level rule. 
 *    <li> Whitespace tokens are artificially set to the special "skip" token
 *         type, which drops them from the token stream seen by the parser.
 *         This is an important feature that simplifies the parser design.
 * </ul>
 * <p>
 * All token types are defined as integer constants in {@link
 * HQLParserTokenTypes} which is an interface that the parser, lexer and
 * other related classes all implement.  This allows all of these classes to
 * directly refer to the token types and share this common set of
 * definitions.  All top-level lexer rules generate a token of the same name
 * in the <code>HQLParserTokenTypes</code> interface.
 * <p> 
 * There is a tokens { } section in the parser where artificial tokens are
 * defined (tokens that are not backed by a lexer rule).  These tokens are
 * also added to the <code>HQLParserTokenTypes</code> interface and can thus
 * be referenced directly by the lexer and parser.
 *
 * @author ECF
 * @author GES
 */
class HQLLexer extends Lexer;

options
{
   k = 3;                                 // lookahead depth is 3 chars
   charVocabulary = '\3'..'\377';         // specifies range 0x03-0xFF
   testLiterals = false;                  // do our own keyword processing
   caseSensitive = false;                 // be case-insensitive
   caseSensitiveLiterals = false;   
}

// helper members and methods
{
   /** Map of keywords to symbol tokens */
   private static Map keywords = new HashMap();

   // Static initializer which populates a keyword map where the symbol text
   // (in the mSYMBOL method) can be submitted to the keyword map for a
   // lookup. If a match is found, then the token type of the token is set
   // based on the token type field stored in the map. This token type will
   // override the default token type of SYMBOL.  Since all of these keywords
   // are not encoded in the matching rules of the grammar itself, there are
   // no actual token definitions in the lexer that match these keywords.
   static
   {
      keywords.put( "and",    new Integer(AND) );
      keywords.put( "case",   new Integer(CASE) );
      keywords.put( "else",   new Integer(ELSE) );
      keywords.put( "end",    new Integer(END) );
      keywords.put( "false",  new Integer(BOOL_FALSE) );
      keywords.put( "is",     new Integer(IS) );
      keywords.put( "like",   new Integer(LIKE) );
      keywords.put( "no",     new Integer(BOOL_FALSE) );
      keywords.put( "not",    new Integer(NOT) );
      keywords.put( "null",   new Integer(NULL) );
      keywords.put( "or",     new Integer(OR) );
      keywords.put( "then",   new Integer(THEN) );
      keywords.put( "true",   new Integer(BOOL_TRUE) );
      keywords.put( "when",   new Integer(WHEN) );
      keywords.put( "yes",    new Integer(BOOL_TRUE) );
      keywords.put( "select", new Integer(SELECT) );
      keywords.put( "from",   new Integer(FROM) );
      keywords.put( "where",  new Integer(WHERE) );
      keywords.put( "as",     new Integer(AS) );
   }
   
   /**
    * Provides a command line interface for an end user to drive and/or test
    * the HQLLexer class.
    * <p>
    * Syntax:
    * <pre>
    *    java HQLLexer &lt;expression&gt;
    * </pre>
    *
    * @param   args 
    *          List of command line arguments.
    */
   public static void main(String[] args)  
   {
      if (args.length != 1)
      {
         System.err.println("Syntax: java HQLLexer <expression>");
         System.exit(-1);
      }
         
      try
      {
         StringReader    expr       = new StringReader(args[0]);
         HQLLexer        lexer      = new HQLLexer(expr);
         int             maxSymSize = 20;
         NumberFormat    lfmt       = NumberFormat.getInstance();
         NumberFormat    cfmt       = NumberFormat.getInstance();
         
         CommonToken     next;
         String          decode;
         
         lfmt.setMinimumIntegerDigits(5);
         lfmt.setGroupingUsed(false);
         cfmt.setMinimumIntegerDigits(3);
         cfmt.setGroupingUsed(false);            
         
         do
         {
            next = (CommonToken) lexer.nextToken();
            
            decode = HQLParser._tokenNames[next.getType()];
            
            StringBuffer sb = new StringBuffer(decode).append('>');
            int trailing = maxSymSize - decode.length();
            
            for (int i = 0; i < trailing; i++)
            {
               sb.append(' ');
            }
            
            System.out.println("["                               +
                               lfmt.format( next.getLine() )     +
                               ":"                               +
                               cfmt.format( next.getColumn() )   +
                               "] <"                             +
                               sb.toString()                     +
                               " "                               +
                               next.getText());
         } 
         while (next.getType() != Token.EOF_TYPE);
      }
       
      catch(Exception excpt)
      {
         excpt.printStackTrace();
      }
   }   
}

/**
 * Matches any amount of whitespace in an expression and sets the token type
 * to SKIP.  In addition, each newline character causes the lexer's line
 * counter to be incremented in order to properly maintain each token's line
 * information.
 * <p>
 * Spaces, tabs and carriage returns and line feeds (newlines) are all
 * matched.
 * <p>
 * This is a top level lexer rule which means that there is an associated
 * <code>WS</code> token.
 */
WS
   :
      ( ' ' | '\t' | '\r' | '\n' { newline(); } )+
      { $setType( Token.SKIP ); }
   ;

/**
 * Match a valid symbol name. All symbols must start with an alphabetic
 * character or the "_" character or the "$" character.  Subsequent
 * characters can be alphanumeric or one of the special symbol characters
 * (<code>_ $</code>).  
 * <b>Symbols are matched case-insensitively.</b>
 * <p>
 * Once a symbol has been found, a keyword lookup occurs. If matched, the
 * the token's type is overridden from the default (<code>SYMBOL</code>)
 * to the artificial token type associated with the keyword.
 * <p>
 * This is a top level lexer rule which means that there is an associated 
 * <code>SYMBOL</code> token.
 */
SYMBOL
   :
      LETTER (VALID_SYM_CHAR)*      
      {
         int stype = SYMBOL;
         
         Integer found = (Integer) keywords.get(getText());
         
         if (found != null)
         {
            // Symbol is a keyword; set the token type accordingly.
            stype = found.intValue();
         }
         
         $setType(stype);
      }
   ;
   
/**
 * Matches the '||' concatenation operator.  This is a top level rule,
 * creating a <code>CONCAT</code> token type.
 */
CONCAT   : "||" ;

/**
 * Matches the ',' character. This is a top level rule, creating a
 * <code>COMMA</code> token type.
 */
COMMA    : ',' ;

/**
 * Matches the "=" string. This is a top level rule, creating an
 * <code>EQUALS</code> token type.
 */
EQUALS   : "=" ;

/**
 * Matches the inequality string. This is a top level rule, creating a
 * <code>NOT_EQ</code> token type.
 */
NOT_EQ   : "!=" ;

/**
 * Matches the '&gt;' character. This is a top level rule, creating a
 * <code>GT</code> token type.
 */
GT       : '>' ;

/**
 * Matches the '&lt;' character. This is a top level rule, creating a
 * <code>LT</code> token type.
 */
LT       : '<' ;

/**
 * Matches the '&gt;=' character sequence. This is a top level rule, creating
 * a <code>GTE</code> token type.
 */
GTE      : ">=" ;

/**
 * Matches the '&lt;=' character sequence. This is a top level rule, creating
 * a <code>LTE</code> token type.
 */
LTE      : "<=" ;

/**
 * Matches the '[' character. This is a top level rule, creating a
 * <code>LBRACKET</code> token type.
 */
LBRACKET  : '[' ;

/**
 * Matches the ']' character. This is a top level rule, creating a
 * <code>RBRACKET</code> token type.
 */
RBRACKET  : ']' ;

/**
 * Matches the '(' character. This is a top level rule, creating a
 * <code>LPARENS</code> token type.
 */
LPARENS  : '(' ;

/**
 * Matches the ')' character. This is a top level rule, creating a
 * <code>RPARENS</code> token type.
 */
RPARENS  : ')' ;

/**
 * Matches the '+' character. This is a top level rule, creating a
 * <code>PLUS</code> token type.
 */
PLUS     : '+' ;

/**
 * Matches the '-' character. This is a top level rule, creating a
 * <code>MINUS</code> token type.
 */
MINUS    : '-' ;

/**
 * Matches the '*' character. This is a top level rule, creating a
 * <code>MULTIPLY</code> token type.
 */
MULTIPLY : '*' ;

/**
 * Matches the '/' character. This is a top level rule, creating a
 * <code>DIVIDE</code> token type.
 */
DIVIDE   : '/' ;

/**
 * Matches the query substitution parameter placeholder character.  This is
 * a question mark (<code>?</code>).  This is a top level rule, creating a
 * <code>SUBST</code> token type.
 */
SUBST    : '?' ;

/**
 * Matches all forms of valid integer literals and decimal literals.
 * Scenarios that are matched:
 * <ul>
 *    <li> Matches a sequence of one or more digits (without a decimal point
 *         and additional digits) as an integer literal with a
 *         <code>NUM_LITERAL</code> token type.
 *    <li> If the literal at previous step can be parsed as integer but cannot fit in 32-bit
 *         space, it will be set to LONG_LITERAL.
 *    <li> If the <code>NUM_LITERAL</code> literal cannot be parsed by <code>Long.parseLong</code>
 *         then it is too large to fit in a 64-bit variable so it will be matched as 
 *         <code>DEC_LITERAL</code>.
 *    <li> Matches a sequence of one or more digits followed by a '.' but
 *         no additional digits, as a decimal literal with a
 *         <code>DEC_LITERAL</code> token type. 
 *    <li> Matches one or more numeric digits followed by a decimal point
 *         and another sequence of numeric digits as a decimal constant 
 *         with a <code>DEC_LITERAL</code> token type.
 *    <li> Matches a decimal point followed by a sequence of numeric digits
 *         as a decimal constant with a <code>DEC_LITERAL</code> token type.
 * </ul>
 * <p>
 * This is a top level rule.  The token type defaults to
 * <code>NUM_LITERAL</code> and is overridden by specific actions depending
 * on the scenario matched.
 */
NUM_LITERAL
   :
      {
         int ntype = NUM_LITERAL;
      }
      (  
           '.' (DIGIT)+              { ntype = DEC_LITERAL; }
         | '.'                       { ntype = DOT;         }
         | (DIGIT)+ ( '.' (DIGIT)*   { ntype = DEC_LITERAL; } )?
         
      )
       {
         if (ntype == NUM_LITERAL)
         {
            try
            {
               long l = Long.parseLong(getText());
               if (l < Integer.MIN_VALUE || l > Integer.MAX_VALUE)
               {
                  ntype = LONG_LITERAL;
               }
            }
            catch (NumberFormatException exc)
            {
               ntype = DEC_LITERAL;
            }
         }
         $setType( ntype );
      }
   ;
   
/**
 * Matches an opening single quote, arbitrary contents and an ending single
 * quote.  Two continguous single quote characters are used to specify a
 * one single quote in the output string (it does not terminate the
 * string).
 * <p>
 * Any newlines inside the string are identified and the lexer's internal
 * newline counter is properly maintained.
 * <p>
 * The greedy option does not need to be disabled here as the closure rule
 * termination is built into the subrule itself: it accepts anything that
 * isn't an unescaped single quote character (see above).
 * <p>
 * Tabs and spaces are maintained inside strings. Opening and closing single
 * quote characters are dropped.
 */
STRING
   :  (
         '\''                                 // opening '
              (
                   "\'\'"                     // allow embedded ''
                 | '\n'    { newline(); }     // allow NL
                 | ~'\''                      // anything except closing '
              )*
         '\''                                 // closing '
      )
   ;

/**   
 * Matches any single character that can appear in the 2nd or later position 
 * in a symbol (matches any <code>LETTER, DIGIT or SYM_CHAR</code>). This is
 * simply a helper rule to make references easier.
 */
protected   
VALID_SYM_CHAR
   :
      LETTER | DIGIT | SYM_CHAR
   ;
   
/** Matches any alphabetic character: <code>a - z</code> */   
protected   
LETTER   : 'a'..'z';
   
/** Matches any numeric digit: <code>0 - 9</code> */   
protected
DIGIT    : '0'..'9';

/**
 * Matches all characters that can be made part of a valid user-defined
 * symbol (except alphabetic and numeric characters).  These are:
 * <code>$ _</code>
 */
protected
SYM_CHAR : '_' | '$';
